class Unit:
	def execute(self, name, deps, options): raise NotImplementedError


class SimpleUnit(Unit):
	def __init__(self, execute_fn):
		self.execute_fn = execute_fn

	def execute(self, *args):
		return self.execute_fn(*args)


def to_unit(fn):
	return SimpleUnit(fn)


class Trendsetter:
	"""
	Trendsetter deals with your complicated dependency tree where every
	interface has several implementations, chosen at runtime,
	with not-so-easy instantiation process for each - and does it lazily,
	and with caching (which doesn't prevent you from having several
	dependency trees, each one cached independently).
	"""
	def __init__(self, options={}):
		self.units = {}
		self.unit_deps = {}
		self.options = options
		self.cache = {}

	def register(self, name: str, unit: Unit, deps=[]):
		self.units[name] = unit
		self.unit_deps[name] = deps

	def get(self, name):
		if name in self.cache:
			return self.cache[name]

		deps = {}
		for i in self.unit_deps[name]:
			deps[i] = self.get(i)

		ret = self.units[name].execute(name, deps, self.options)
		self.cache[name] = ret
		return ret


if __name__ == '__main__':
	class Adder:
		def add(self, a, b): raise NotImplementedError

	class AdderImpl(Adder):
		def add(self, a, b):
			return a + b

	class ExperimentalAdderImpl(Adder):
		def add(self, a, b):
			return a  # TODO

	class AdderUnit(Unit):
		def execute(self, name, deps, options):
			if options['use_experimental_adder']:
				return ExperimentalAdderImpl()
			else:
				return AdderImpl()

	ts = Trendsetter({
		# No assumptions what your options look like.
		'use_experimental_adder': False
	})
	ts.register('Adder', AdderUnit())

	class Fibonacci:
		def __init__(self, adder):
			self.adder = adder

		def compute(self, n):
			if n <= 2:
				return 1
			else:
				return self.adder.add(self.compute(n-1), self.compute(n-2))

	ts.register(
		'Fibonacci',
		# Simplified interface for more straightforward cases
		to_unit(lambda name, deps, options: Fibonacci(deps['Adder'])),
		# List of dependencies - they will be instantiated first
		['Adder'],
	)

	assert ts.get('Adder').add(2, 2) == 4
	assert ts.get('Fibonacci').compute(6) == 8
	print('6th Fibonacci number is:', ts.get('Fibonacci').compute(6))



from .date import Week, get_week, WeekNumbering

import pytest
from arrow import get as arrow_get


GET_WEEK_TEST_DATA = [
	(arrow_get(2018, 12, 30), WeekNumbering.MATCH_YEAR, Week(52, 2018)),
	(arrow_get(2018, 12, 31), WeekNumbering.MATCH_YEAR, Week(53, 2018)),
	(arrow_get(2019, 1, 1), WeekNumbering.MATCH_YEAR, Week(0, 2019)),
	(arrow_get(2019, 1, 7), WeekNumbering.MATCH_YEAR, Week(1, 2019)),
	(arrow_get(2021, 12, 31), WeekNumbering.MATCH_YEAR, Week(52, 2021)),
	(arrow_get(2022, 1, 1), WeekNumbering.MATCH_YEAR, Week(0, 2022)),
	(arrow_get(2022, 12, 31), WeekNumbering.MATCH_YEAR, Week(52, 2022)),
	(arrow_get(2023, 1, 1), WeekNumbering.MATCH_YEAR, Week(0, 2023)),
	(arrow_get(2023, 1, 2), WeekNumbering.MATCH_YEAR, Week(1, 2023)),
	(arrow_get(2023, 1, 3), WeekNumbering.MATCH_YEAR, Week(1, 2023)),
	(arrow_get(2023, 12, 31), WeekNumbering.MATCH_YEAR, Week(52, 2023)),
	(arrow_get(2024, 1, 1), WeekNumbering.MATCH_YEAR, Week(1, 2024)),
	(arrow_get(2024, 1, 8), WeekNumbering.MATCH_YEAR, Week(2, 2024)),
	(arrow_get(2018, 12, 30), WeekNumbering.NORMAL, Week(52, 2018)),
	(arrow_get(2018, 12, 31), WeekNumbering.NORMAL, Week(53, 2018)),
	(arrow_get(2019, 1, 1), WeekNumbering.NORMAL, Week(53, 2018)),
	(arrow_get(2019, 1, 7), WeekNumbering.NORMAL, Week(1, 2019)),
	(arrow_get(2021, 12, 31), WeekNumbering.NORMAL, Week(52, 2021)),
	(arrow_get(2022, 1, 1), WeekNumbering.NORMAL, Week(52, 2021)),
	(arrow_get(2022, 12, 31), WeekNumbering.NORMAL, Week(52, 2022)),
	(arrow_get(2023, 1, 1), WeekNumbering.NORMAL, Week(52, 2022)),
	(arrow_get(2023, 1, 2), WeekNumbering.NORMAL, Week(1, 2023)),
	(arrow_get(2023, 1, 3), WeekNumbering.NORMAL, Week(1, 2023)),
	(arrow_get(2023, 12, 31), WeekNumbering.NORMAL, Week(52, 2023)),
	(arrow_get(2024, 1, 1), WeekNumbering.NORMAL, Week(1, 2024)),
	(arrow_get(2024, 1, 8), WeekNumbering.NORMAL, Week(2, 2024)),
]


@pytest.mark.parametrize(
	"date, mode, expected_week_number",
	GET_WEEK_TEST_DATA,
)
def test_get_week(date, mode, expected_week_number):
	assert get_week(date, mode) == expected_week_number



from .base import Query

from sqlalchemy.sql import text


def get_mutations_in_progress_count(db_name: str, table_name: str) -> int:
	"""
	Requests from the System.Mutants table the number of mutations in progress.

	Parameters
	----------
	db_name: str
		Must contain name of existed database with tables.
	table_name: str
		Must contain name of existed table in database with db_name.

	Returns
	-------
	int
		Number of mutants in progress.
	"""

	# TODO: Find a way to use FROM construction. By default, requests are redirected to the DB specified during
	#  connection
	# query = Query(
	# 	select([func.count(System_Mutations)])
	#
	# 	.where('database' == db_name)
	# 	.where('table' == 'order')
	# 	.where('is_done' == 0)
	# 	.group_by('is_done')
	# )

	query = Query()
	query.q = text("""SELECT count(*)
FROM system.mutations
WHERE database == :db_name AND table == :table_name AND is_done == 0""")

	query.params = {'db_name': db_name, 'table_name': table_name}

	return query.execute_val()



class Combinator:
	"""
	Combines multiple data sets into a set of their combinations.
	Say, our datasets are [1,2], [3,4], [5,6].
	Then, our combinations would be:
	with 1 dataset:  [[1], [2]]
	with 2 datasets: [[1,3], [1,4], [2,3], [2,4]]
	with 3 datasets: [[1,3,5], [1,3,6], [1,4,5], [1,4,6], [2,3,5], [2,3,6], [2,4,5], [2,4,6]]
	"""
	result = []

	def __init__(self):
		self.result = []

	def combine(self, new_set):
		#print("Current result", self.result)
		#print("Combining set", new_set)

		if len(new_set) == 0:
			raise Exception("Combinator cannot combine with empty set")

		if len(self.result) == 0:
			for i in new_set:
				self.result += [[i]]
		else:
			newresult = []
			for tuple in self.result:
				for i in new_set:
					newresult += [tuple + [i]]
			self.result = newresult
		#print("New result", self.result)

	def get_result(self):
		return self.result



from .mutations import get_mutations_in_progress_count

from time import sleep
from typing import Union


def wait_result(db_name: str, table_name: str, delay: Union[int, float] = 0.25) -> None:
	"""
	Suspends the program while there are incomplete mutant processes.

	Parameters
	----------
	db_name: str
		Database name in which the deletion operation was performed.
	table_name: str
		Table name in the {db_name} database in which the deletion operation was performed.
	delay: Union[int, float]
		Delay between system table polls.

	Notes
	-----
	Imagine that you are driving in a car to a railway crossing. A train is going to cross you by rail.
	This function is a handbrake that will save you from crashing.
	"""

	while get_mutations_in_progress_count(db_name, table_name):
		sleep(delay)



from .filterset import create_filter_set, from_request

import pytest
from django.forms import IntegerField, CharField


def test_filter_set():
	CountryFilterSet = create_filter_set({
		'country': CharField(required=True),
	}, required=['country'])

	AddressFilterSet = create_filter_set({
		'city': CharField(required=True),
		'street': CharField(),
		'building': IntegerField(),
	}, required=['city'], parent=CountryFilterSet)

	with pytest.raises(AssertionError):
		from_request(AddressFilterSet, {})

	with pytest.raises(AssertionError):
		from_request(AddressFilterSet, {
			'country': 'Russia',
		})

	source = {
		'country': 'Russia',
		'city': 'Penza',
		'street': 'Hi st.',
		'building': '344',
	}
	expected = {
		'country': 'Russia',
		'city': 'Penza',
		'street': 'Hi st.',
		'building': 344,
	}
	r = from_request(AddressFilterSet, source)
	assert r.is_valid
	r.data.unfreeze(['country', 'city', 'street', 'building'])
	assert r.data.as_dict() == expected



from pandas import DataFrame, notnull


def row_to_dict(d, key_as=None):
	key, value = d
	ret = value.to_dict()
	if key_as is not None:
		ret[key_as] = key
	return ret


def replace_null_objects(dataframe: DataFrame, new_value=None) -> DataFrame:
	"""
	Returns pandas DataFrame with changed (pd.NaT, pd.NaN and NULL) values to <new_value>

	Parameters
	----------
	dataframe: DataFrame
		Source dataframe in which will be changed all null values.
	new_value: Any | Default None
		New null value object.

	Returns
	-------
	DataFrame:
		Pandas dataframe with replaced all null values to new_value.

	Notes
	-----
	Sometimes (especially in older versions of pandas) it happens that when importing a file, columns are assigned
	types, and empty cells are converted to corresponding null types. To standardize the format of null values or change
	it to None, I recommend using this function.
	"""

	dataframe = dataframe.where(notnull(dataframe), new_value)

	return dataframe



import json
from typing import Iterable, Dict, Hashable

from django.http import StreamingHttpResponse


def stream_json(data: Iterable[Dict[Hashable, any]]) -> StreamingHttpResponse:
	"""
	Stream all elements in `data` as JSON array using the StreamingHttpResponse class.

	Parameters
	----------
	data: Iterable[Dict[Hashable, any]]
		An Iterable of JSON-encodable elements.

	Returns
	-------
	StreamingHttpResponse
		A streaming HTTP response class with an iterator as content. Content-type = 'application/json'

	Notes
	-----
	The function is recommended to be used with a large set of transmitted data.
	"""

	assert isinstance(data, Iterable), "stream_json: data must be of iterable type"

	def _iter(ret):
		first = True
		yield '['

		for x in ret:
			if first:
				first = False
				yield json.dumps(x, ensure_ascii=False)
			else:
				yield ', '
				yield json.dumps(x, ensure_ascii=False)
		yield ']'

	return StreamingHttpResponse(
		_iter(data),
		content_type="application/json",
	)



from .views import DirtyHackAuthCheckViewSet, auth_view

from django.urls import path, include
from rest_framework import routers


router = routers.DefaultRouter()
router.register(r'check', DirtyHackAuthCheckViewSet, basename='xauth_check')

urlpatterns = [
	path('auth-post', auth_view),
	path('', include(router.urls)),
]



from typing import Sequence, Hashable


def find_field(d, candidates):
	"""
	Given a dict and a list of possible keys, find the first key
	which is included into this dict. Throws ValueError if not found.
	"""
	for c in candidates:
		if c in d:
			return c
	raise ValueError(f"Can't find any of: {candidates}")

def find_value(d, candidates):
	"""
	Given a dict and a list of possible keys, find the first key
	which is included into this dict and return its value.
	Throws ValueError if not found.
	"""
	fieldname = find_field(d, candidates)
	return d[fieldname]

def normalize_dict(source, replacements, allow_original_key=True):
	"""
	Given a dict and a map: reference key -> [] possible replacements,
	return a new dict where all keys are guaranteed to be reference keys
	with values taken from any of replacement keys.

	Typically used to normalize a dataset where the same info can be listed
	under different names.

	Example map:
	{
		"artist": ["Artist", "ARTIST"],
		"year": ["Year", "yr", "yob"],
	}
	"""
	ret = {}
	for key in replacements:
		candidates = replacements[key]
		if allow_original_key:
			candidates += (key,)
		value = find_value(source, candidates)
		ret[key] = value
	return ret


def filter_dict(src: dict, keys_to_filter: Sequence[Hashable]) -> dict:
	"""
	Filters dictionary by keys_to_filter set.

	Parameters
	----------
	src: dict
		Source dictionary.
	keys_to_filter: Sequence[Hashable]
		Set of keys that should be in the final dictionary.

	Returns
	-------
	dict
		Filtered source dictionary.
	"""

	return {key: value for key, value in src.items() if key in keys_to_filter}



from ..date import parse_month
from django import forms


class MonthField(forms.CharField):
	def clean(self, value):
		if value is None: return None
		try:
			return parse_month(value)
		except:
			raise forms.ValidationError('Not a valid month (expected YYYY-MM)')


class ChoiceWithDefault(forms.ChoiceField):
	def __init__(self, *args, **kwargs):
		default_value = kwargs['default']
		del kwargs['default']
		super(ChoiceWithDefault, self).__init__(*args, **kwargs)
		self.default_value = default_value

	def clean(self, value):
		if value is None:
			return self.default_value
		return super(ChoiceWithDefault, self).clean(value)


class IntegerArrayField(forms.CharField):
	def __init__(self, sep='|', *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.sep = sep

	def clean(self, value):
		if value is None:
			return None
		return list(map(int, value.split(self.sep)))


class StringArrayField(forms.CharField):
	def __init__(self, sep='|', strip=True, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.sep = sep
		self.strip = strip

	def clean(self, value):
		if value is None:
			return None
		ret = value.split(self.sep)
		if self.strip:
			ret = list(map(lambda value: value.strip(), ret))
		return ret


class MonthArrayField(forms.CharField):
	def __init__(self, sep='|', *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.sep = sep

	def clean(self, value):
		if value is None: return None
		value = value.split('|')
		try:
			return list(map(lambda value: parse_month(value), value))
		except:
			raise forms.ValidationError('Not a valid month (expected YYYY-MM)')



class Timing(object):
	"""
	Measures time that an operation took to complete.

		timing = Timing(auto_print=True)
		with timing:
			do_something()

		timing = Timing()
		with timing:
			do_something()
		print(timing.result)
	"""
	def __init__(self, auto_print=False):
		super(Timing, self).__init__()
		self.result = None
		self._start = None
		self.auto_print = auto_print

	def __enter__(self):
		import arrow
		self._start = arrow.now()

	def __exit__(self, exc_type, exc_value, traceback):
		import arrow
		self.result = arrow.now() - self._start
		if self.auto_print:
			print('time:', self.result)



from .formplus import MonthArrayField

from collections import namedtuple

from django.core.exceptions import ValidationError
import pytest
from hypothesis import given
from hypothesis.strategies import lists, dates


def test_MonthArrayField():
	return
	TestCase = namedtuple('TestCase', ['input', 'output', 'valid'])
	data = [
		TestCase(input='2019-01', output=[(2019, 1)], valid=True),
		TestCase(input='1999-02', output=[(1999, 2)], valid=True),
		TestCase(input='2019-01|2019-02|1917-10', output=[(2019, 1), (2019, 2), (1917, 10)], valid=True),
		TestCase(input='2678', valid=False, output=None),
		TestCase(input='-01', valid=False, output=None),
		TestCase(input='2019-01|672643787878', valid=False, output=None),
		TestCase(input='2019-01|2099-01||||||||||||||||	  ', valid=False, output=None),
		TestCase(input='2019-13', output=None, valid=False),
	]

	for d in data:
		f = MonthArrayField()
		if d.valid:
			assert f.clean(d.input) == d.output
		else:
			with pytest.raises(ValidationError):
				f.clean(d.input)


@given(lists(dates(), min_size=1))
def test_MonthArrayField_correctness(source):
	source_str = '|'.join(map(lambda x: '-'.join(str(x).split('-')[:2]), source))
	f = MonthArrayField()
	result = f.clean(source_str)
	for i in range(0, len(result)):
		assert result[i][0] == source[i].year
		assert result[i][1] == source[i].month



from .array import ensure_iterable
from .deprecated import deprecated

import datetime
from typing import Union, NamedTuple
from abc import ABC, abstractmethod
from enum import Enum, auto, unique

import arrow


DateLikeObject = Union[datetime.date, datetime.datetime, arrow.Arrow]


class Week(NamedTuple):
	week: int
	year: int

	def __str__(self) -> str:
		return f"{self.week} week of {self.year} year"


@unique
class WeekNumbering(Enum):
	"""
	Contains enumeration of available modes of week extraction.
	"""

	"""
	First day in week is: Monday
	First week number in year is: One
	First week is a week with: The first of Monday
	"""
	NORMAL = auto()

	"""
	First day in week is: Monday
	First week number in year: Zero
	First week is a week with: The first of January
	"""
	MATCH_YEAR = auto()


def parse_month(yyyy_mm):
	a = arrow.get(yyyy_mm, 'YYYY-M')
	return (a.year, a.month)


def parse_date(s, format_or_formats):
	d = None
	for f in ensure_iterable(format_or_formats):
		try:
			d = arrow.get(s, f)
			break
		except:  # should probably narrow to arrow.ParserError
			pass
	if d == None:
		raise ValueError(f"Could not parse date `{s}` with any of the specified formats")
	return d


def normalize_date(s, input_formats, output_format, allow_empty=True):
	if s == None and allow_empty:
		return None
	return parse_date(s, input_formats).format(output_format)


class WeekExtractor(ABC):
	@classmethod
	@abstractmethod
	def extract(cls, date_object: DateLikeObject) -> Week: ...


class MatchYearWeekExtractor(WeekExtractor):
	@classmethod
	def extract(cls, date_object: DateLikeObject) -> Week:
		"""
		Returns week number of specified date with year.
		First week of year is week with monday.
		Range of week numbers: 0..53.
		0 means that week refers to the last week of the previous year.
		"""
		return Week(int(date_object.strftime('%W')), date_object.year)


class NormalWeekExtractor(WeekExtractor):
	@classmethod
	def extract(cls, date_object: DateLikeObject) -> Week:
		"""
		Returns week number of specified date with year.
		First week of year is week with monday.
		Range of week numbers: 1..53.
		"""
		week = MatchYearWeekExtractor.extract(date_object)

		if week.week == 0:
			# Means that the week refers to the last week of the previous year
			week = MatchYearWeekExtractor.extract(datetime.date(date_object.year - 1, 12, 31))

		return week


@deprecated('notalib: current get_week behavior is considered buggy and will be changed in 2.0.')
def get_week(
	date_object: DateLikeObject,
	mode: WeekNumbering = WeekNumbering.NORMAL,
) -> Week:
	"""
	Extracts week from specified date.

	Args:
		date_object: Date like object. It must be compatible with 'datetime.date' object.
		mode: Mode of extraction week number. Available modes see in WeekNumbering.
	"""
	if not isinstance(mode, WeekNumbering):
		raise TypeError(f"The mode {mode} does not belong to the {WeekNumbering.__name__} enumeration")

	if mode == WeekNumbering.NORMAL:
		return NormalWeekExtractor.extract(date_object)
	elif mode == WeekNumbering.MATCH_YEAR:
		return MatchYearWeekExtractor.extract(date_object)

	raise NotImplementedError(f'Extraction week number in mode {mode.name} is not implemented')



import logging
from blessings import Terminal; t = Terminal()
from datetime import datetime


class ColorFormatter(logging.Formatter):
	"""
	Example usage:
	
	LOGGING = {
		'version': 1,
		'disable_existing_loggers': False,
		'formatters': {
			'colors': {
				'()': 'util.colorlog.ColorFormatter',
			},
			'stamp': {
				'format': '%(asctime)s %(name)s [%(levelname)s] %(message)s',
				# 'format': '%(asctime)s %(name)s %(module)s [%(levelname)s] %(message)s',
			},
		},
		'handlers': {
			'console': {
				'class': 'logging.StreamHandler',
				'formatter': 'colors',
			},
		},
		'loggers': {
			# 'django': {
			# 	'handlers': ['console'],
			# 	'level': 'DEBUG',
			# },
			'dialer': {
				'handlers': ['console'],
				'level': 'DEBUG',
			},
		},
	}
	"""
	
	def __init__(self, *args, **kwargs):
		super(ColorFormatter, self).__init__()

	def format(self, record):
		lvl = record.levelname
		if record.levelname == 'DEBUG': lvl = t.white('DBG')
		if record.levelname == 'INFO': lvl = t.blue('INF')
		if record.levelname == 'WARNING': lvl = t.yellow('WRN')
		if record.levelname == 'ERROR': lvl = t.red('ERR')
		if record.levelname == 'CRITICAL': lvl = t.bright_red('CRT')

		NAME_LENGTH_MAX = 19
		name = record.name
		if len(name) > NAME_LENGTH_MAX:
			name = name[-NAME_LENGTH_MAX:]

		return f'{datetime.strftime(datetime.now(), "%H:%M:%S.%f")}|{name}[{lvl}] {record.getMessage()}'



import time
import logging; log = logging.getLogger(__name__)

from django.conf import settings


CLICKHOUSE_PROFILE = settings.CLICKHOUSE_PROFILE


########
# How to use:
#
# from .clickhouse_profiler import before_cursor_execute, after_cursor_execute
# ...
# event.listens_for(Engine, "before_cursor_execute")(before_cursor_execute)
# event.listens_for(Engine, "after_cursor_execute")(after_cursor_execute)
########


def indent(n: int, s: str):
	padding = ' '*n
	return '\n'.join(map(lambda x: padding+x, s.split('\n')))


def midrange_cut(s: str, n: int):
	if len(s) <= n:
		return s
	right = n // 10
	left = n - right
	ret = s[:left] + '\n...\n' + s[-right:]
	return ret


#@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
	if not CLICKHOUSE_PROFILE: return
	context._query_start_time = time.time()


#@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
	if not CLICKHOUSE_PROFILE: return
	total = time.time() - context._query_start_time

	parameters_s = midrange_cut(str(parameters), 200)
	statement_s = midrange_cut(statement, 500)

	s = f'''
Query:  {indent(8, statement_s).strip()}
params: {indent(8, parameters_s).strip()}
-- time: {"%.02fms" % (total*1000)} ------------------------------
'''
	print(s)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['notalib',
 'notalib.django',
 'notalib.django.clickhouse',
 'notalib.django_xauth',
 'notalib.django_xauth.migrations',
 'notalib.pandas']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.14.0,<2.0.0']

setup_kwargs = {
    'name': 'notalib',
    'version': '2.0.0',
    'description': 'A collection of utility functions & classes',
    'long_description': '# notalib [![Django CI](https://github.com/m1kc/notalib/actions/workflows/django.yml/badge.svg)](https://github.com/m1kc/notalib/actions/workflows/django.yml) [![Coverage Status](https://coveralls.io/repos/github/m1kc/notalib/badge.svg?branch=master)](https://coveralls.io/github/m1kc/notalib?branch=master)\n\nCollection of small Python utility functions and classes. Each one was created because I needed it and it didn\'t exist or I didn\'t like the existing implementations. 100% of code is used in real-world projects.\n\n(And one day, the documentation is going to be actually good. In the meanwhile, don\'t hesitate to ask if something\'s not clear.)\n\n## Install\n\n```sh\npip install notalib\n```\n\nOr with [poetry](https://python-poetry.org/):\n\n```sh\npoetry add notalib\n```\n\n## Maintenance & bugfixes\n\nWhile I try to fix bugs, add new features, and review any PRs when I have time, there\'re no promises and no set timeframes, even if a bug is critical. That\'s a project I do in my free time, free of charge.\n\nIf that\'s not enough for you or you have an urgent request, there are paid maintenance options (bugfixing, features, expedite PR review, 24h security responses). Contact me for prices: m1kc@yandex.ru\n\nAlso feel free to just send me money:\n\n* MasterCard: 5559 4925 7484 0297\n* PayPal: [paypal.me/thisism1kc](https://paypal.me/thisism1kc)\n\nDonations are always appreciated, even if you send 10$.\n\n## Constants included\n\n* `notalib.utf.BOM`: contains string `b\'\\xEF\\xBB\\xBF\'` (UTF-8 little endian byte order mark).\n\n## Utils included\n\n#### notalib.array.as_chunks :fire:\n\nIterates over your array in chunks of at most N elements.\n\n```python\nfrom notalib.array import as_chunks\n\narr = [1,2,3,4,5]\nfor chunk in as_chunks(arr, 2):\n    print(chunk)\n# [1,2]\n# [3,4]\n# [5]\n```\n\n#### notalib.array.ensure_iterable :fire:\n\nKeeps iterable things like lists intact, turns single values into single-element lists. Useful for functions that can accept both.\n\n```python\nensure_iterable([1,2,3])  # --> [1,2,3]\nensure_iterable((1,2,3))  # --> (1,2,3)\nensure_iterable(1)        # --> [1]\nensure_iterable(\'smth\')   # --> [\'smth\']\n\ndef my_function(one_or_multiple_args):\n    for arg in ensure_iterable(one_or_multiple_args):\n        ...\n        \nmy_function([\'log\', \'smog\'])\nmy_function(\'dog\')\n```\n\n#### notalib.combinator.Combinator :fire:\n#### notalib.date.parse_month\n#### notalib.date.parse_date\n#### notalib.date.normalize_date :fire:\n#### <s>notalib.date.get_week_number</s>\n\n_Removed in 2.0.0. Use `get_week` instead. If you want the "old" week numbering, use get_week with `WeekNumbering.MATCH_YEAR` and add 1 to week number._\n\n#### notalib.date.get_week\n\nReturns named tuple with week number for the given date. Accepts Python dates and Arrow timestamps.\n\nOptional argument `mode` tells what to do if the week started in previous year:\n\n* WeekNumbering.NORMAL (default): consider it the last week of the previous year\n* WeekNumbering.MATCH_YEAR: consider it 0-th week of current year\n\n```python\nfrom notalib.date import get_week, WeekNumbering\nfrom datetime import date\n\ndate1, date2 = date(2022, 12, 31), date(2023, 1, 1)\nget_week(date1, WeekNumbering.NORMAL)\n# Week(week=52, year=2022)\n\nget_week(date1, WeekNumbering.MATCH_YEAR)\n# Week(week=52, year=2022)\n\nget_week(date2, WeekNumbering.NORMAL)\n# Week(week=52, year=2022)\n\nget_week(date2, WeekNumbering.MATCH_YEAR)\n# Week(week=0, year=2023)\n```\n\n#### notalib.dict.find_field\n#### notalib.dict.find_value\n#### notalib.dict.normalize_dict :fire:\n#### notalib.dict.filter_dict\n\nFilters a dictionary, removing any keys except for the ones you choose.\n\n```python\nfrom notalib.dict import filter_dict\n\nsrc = {\n\t\'Some\': "BODY",\n\t\'once\': "told me",\n\t\'the world\': "is gonna roll me",\n}\nfilter_dict(src, ["Some", "once"])\n# {\'Some\': \'BODY\', \'once\': \'told me\'}\nfilter_dict(src, [])\n# {}\n```\n\n#### notalib.format.format_long_list\n#### notalib.hypertext.strip_tags :fire:\n#### notalib.hypertext.TablePrinter :fire:\n\nPrints an HTML table, row by row, from the given data, using attrs or dictionary keys as columns.\n\nTwo ways to use it:\n\n* Call header() / entry() / footer() manually\n\n```python\nfrom notalib.hypertext import TablePrinter\nt = TablePrinter([\'a\', \'b\'])\nt.header()\n# \'<table><thead><tr><th>a</th><th>b</th></tr></thead><tbody>\'\nt.entry({\'a\': 1, \'b\': 2})\n# \'<tr><td>1</td><td>2</td></tr>\\n\'\nt.entry({\'a\': 11, \'b\': 22})\n# \'<tr><td>11</td><td>22</td></tr>\\n\'\nt.footer()\n# \'</tbody></table>\'\n```\n\n* Pass an iterable to iterator_over()\n\t\n```python\nfrom notalib.hypertext import TablePrinter\nt = TablePrinter([\'a\', \'b\'])\nlist(t.iterator_over([ {\'a\': 11, \'b\': 22} ]))\n# [\'<table><thead><tr><th>a</th><th>b</th></tr></thead><tbody>\',\n#  \'<tr><td>11</td><td>22</td></tr>\\n\',\n#  \'</tbody></table>\']\n```\n\n#### notalib.polosa.polosa :fire: :fire: :fire: :fire: :fire:\n\n```\n18023/2000000   294.8/sec   Processing transaction ID#84378473 (2020-01-04)\n```\n\nThe CLI progress indicator you\'ve always dreamt of: shows current and total if available, measures current speed, can show your comments for each element, makes sure not to slow down your terminal with frequent updates. [See this short demo](https://asciinema.org/a/UI1aOqjQC1KXx303kaVGrxjQp).\n\n_Cheat sheet_\n\n```python\n## Basic usage\nwith polosa() as p:\n    p.tick()\n# 467344   201.2/sec\n\n## Specify total number of elements:\nwith polosa(total=1337) as p:\n# 26/1337   1.2/sec\n\n## Print something useful about every element:\np.tick(caption=my_order.time_created)\n# 1723910/2000000   319231.2/sec   2020-01-01 15:37:00\n```\n\n#### notalib.range.Range\n#### notalib.time.Timing :fire:\n\nMeasures time spent on executing your code. Killer feature: it can be used as a reusable context.\n\n```python\ntiming = Timing()\n...\nwith timing:\n    do_something()\n# That\'s it, do something with the measurement\nlog(f\'Operation took {timing.result} sec\')\n```\n\nIf you just want to print measurements into console, there\'s a shorthand:\n\n```python\ntiming = Timing(auto_print=True)\n...\nwith timing:\n    do_something()\n```\n\n#### notalib.trendsetter.Trendsetter :fire:\n\n## Pandas-related\n\n#### notalib.pandas.pandasplus.row_to_dict\n#### notalib.pandas.pandasplus.replace_null_objects\n\nReplaces all types of null values in a DataFrame with the given value.\n\n```python\ndf = pd.DataFrame({\'A\': [pd.NA, pd.NaT, \'SomeVal\', None]})\nnew_df = replace_null_objects(df, "Hello, notalib!")\nnew_df\n#                  A\n# 0  Hello, notalib!\n# 1  Hello, notalib!\n# 2          SomeVal\n# 3  Hello, notalib!\n```\n\n## Django-related\n\n#### notalib.django.auth.StaticBackend\n#### notalib.django.auth.SettingsBackend\n\n#### notalib.django.xauth\n\nEndpoints for easier authentication in APIs. Requires Django REST framework.\n\nProvides endpoints:\n\n* `/xauth/check` — returns code 200 if client is authenticated (or global permissions are set to AllowAny), 403 if not\n* `/xauth/auth-post` — authenticates a client; accepts two POST parameters `username` and `password`; returns code 200 on success and 403 on failure\n\nHow to use:\n\n1. Make sure Django REST framework is installed.\n2. Add `\'notalib.django_xauth\'` to INSTALLED_APPS.\n3. Run `manage.py migrate django_xauth` (doesn\'t actually change your DB).\n4. Add something like this to your urls.py: `path(\'xauth/\', include(\'notalib.django_xauth.urls\')),`\n\n#### notalib.django.colorlog.ColorFormatter\n#### notalib.django.filterset :fire:\n#### notalib.django.formplus.MonthField\n#### notalib.django.formplus.ChoiceWithDefault\n#### notalib.django.formplus.IntegerArrayField\n#### notalib.django.formplus.StringArrayField\n#### notalib.django.formplus.MonthArrayField\n#### notalib.django.request_time_middleware.RequestTimeLoggingMiddleware\n#### notalib.django.stream.stream_json\n\nStream all elements of iterable object as JSON array using the StreamingHttpResponse class. Unlike DRF\'s Response class, it can handle arrays of any size.\n\n```python\nclass SomeViewSet(...):\n    ...\n    \n    def list(self, request, *args, **kwargs):\n        ...\n        return stream_json(data)\n```\n\n## Django/Clickhouse\n\n_Stage: alpha_\n\nRequired packages: `clickhouse-sqlalchemy`\n\nRequired two django.settings variables:\n\n* **CLICKHOUSE_URL** — database URL\n* **CLICKHOUSE_PROFILE** — dump all queries and their timing to console, true/false\n\n#### notalib.django.clickhouse.base.get_connection\n#### notalib.django.clickhouse.base.get_database_name\n#### notalib.django.clickhouse.base.Query\n\nA wrapper for SQLAlchemy\'s `select` with some useful postprocessing options.\n\n* `execute` — no postprocessing\n* `execute_val` — returns single value\n* `execute_list` — returns single column as list\n* `execute_kv` — returns dict, first column becomes keys, second column becomes values\n* `execute_na` — returns number of affected rows\n\nUsage example:\n\n```python\nq = Query(\n    select([ SomeTable.c.notalib ])\n)\nq.execute_list()\n# ["Example", "OOOOO", "my", "defence", ...]\n```\n\n#### notalib.django.clickhouse.mutations.get_mutations_in_progress_count\n\nReturns the number of mutations in progress for the specified table.\n\n```python\nget_mutations_in_progress_count("SOME_DATABASE", "SOME_TABLE_IN_DATABASE")\n# 5\n```\n\n#### notalib.django.clickhouse.wait.wait_result :fire:\n\nWaits until all mutations for the given table are complete.\n\n```python\n# blah blah blah, ALTER TABLE ... UPDATE ...\nwait_result("SOME_DATABASE", "SOME_TABLE_IN_DATABASE")\n# UPDATE complete, continue\n```\n',
    'author': 'm1kc (Max Musatov)',
    'author_email': 'm1kc@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/m1kc/notalib',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.3,<4.0.0',
}


setup(**setup_kwargs)



from typing import Callable, Type
from warnings import warn


def deprecated(
	reason: str = 'Unspecified reason',
	warning: Type[DeprecationWarning] = DeprecationWarning,
	stacklevel: int = 2,
) -> Callable:
	"""
	Marks function or method as deprecated.

	Args:
		reason: The reason why the code has become deprecated.
		warning: Deprecated code warning class.
		stacklevel: Logging stacklevel.

	Note:
		See more about stacklevel in official docs:
			Logging: https://docs.python.org/3/library/logging.html#logging.Logger.debug
			Warnings: https://docs.python.org/3/library/warnings.html#warnings.warn
	"""
	def decorator(func: Callable):
		if not callable(func):
			raise TypeError('The "func" argument must be of Callable type')

		def inner(*args, **kwargs):
			warn(reason, warning, stacklevel=stacklevel)
			return func(*args, **kwargs)

		return inner

	return decorator



import datetime
import uuid

from django.utils.deprecation import MiddlewareMixin

class RequestTimeLoggingMiddleware(MiddlewareMixin):
    """Middleware class logging request time to stderr.

    This class can be used to measure time of request processing
    within Django.  It can be also used to log time spent in
    middleware and in view itself, by putting middleware multiple
    times in INSTALLED_MIDDLEWARE.

    Static method `log_message' may be used independently of the
    middleware itself, outside of it, and even when middleware is not
    listed in INSTALLED_MIDDLEWARE.
    """

    @staticmethod
    def log_message(request, tag, message=''):
        """Log timing message to stderr.

        Logs message about `request' with a `tag' (a string, 10
        characters or less if possible), timing info and optional
        `message'.

        Log format is "timestamp tag uuid count path +delta message"
        - timestamp is microsecond timestamp of message
        - tag is the `tag' parameter
        - uuid is the UUID identifying request
        - count is number of logged message for this request
        - path is request.path
        - delta is timedelta between first logged message
          for this request and current message
        - message is the `message' parameter.
        """

        dt = datetime.datetime.utcnow()
        if not hasattr(request, '_logging_uuid'):
            request._logging_uuid = uuid.uuid1()
            request._logging_start_dt = dt
            request._logging_pass = 0

        request._logging_pass += 1
        print('%s %-10s %s %2d %s +%s %s' % (
            dt.isoformat(),
            tag,
            request._logging_uuid,
            request._logging_pass,
            request.path,
            dt - request._logging_start_dt,
            message,
        ))

    def process_request(self, request):
        self.log_message(request, 'request ')

    def process_response(self, request, response):
        s = getattr(response, 'status_code', 0)
        r = str(s)
        if s in (300, 301, 302, 307):
            r += ' => %s' % response.get('Location', '?')
        elif response.content:
            r += ' (%db)' % len(response.content)
        self.log_message(request, 'response', r)
        return response
